@@ -12,6 +12,15 @@ use std::sync::Arc;
 
 use im_ternary_tree::TernaryTreeList;
 
+/// only macro and func are cared about during preprocessing
+/// only used in preprocess defs
+fn pick_macro_fn(x: Calcit) -> Option<Calcit> {
+  match &x {
+    Calcit::Fn { .. } | Calcit::Macro { .. } => Some(x),
+    _ => None,
+  }
+}
+
 /// returns the resolved symbol,
 /// if code related is not preprocessed, do it internally
 pub fn preprocess_ns_def(
@@ -41,7 +50,7 @@ pub fn preprocess_ns_def(
             rule: import_rule,
           })),
         },
-        Some(v),
+        pick_macro_fn(v),
       ))
     }
     None => {
@@ -84,7 +93,7 @@ pub fn preprocess_ns_def(
                 rule: Some(Arc::new(ImportRule::NsReferDef(ns.to_owned(), def.to_owned()))),
               })),
             },
-            Some(v),
+            pick_macro_fn(v),
           ))
         }
         None if ns.starts_with('|') || ns.starts_with('"') => Ok((
@@ -273,12 +282,10 @@ pub fn preprocess_expr(
         process_list_call(xs, scope_defs, file_ns, check_warnings, call_stack)
       }
     }
-    Calcit::Number(..) | Calcit::Str(..) | Calcit::Nil | Calcit::Bool(..) | Calcit::Keyword(..) => {
-      Ok((expr.to_owned(), Some(expr.to_owned())))
-    }
+    Calcit::Number(..) | Calcit::Str(..) | Calcit::Nil | Calcit::Bool(..) | Calcit::Keyword(..) => Ok((expr.to_owned(), None)),
     Calcit::Proc(..) => {
       // maybe detect method in future
-      Ok((expr.to_owned(), Some(expr.to_owned())))
+      Ok((expr.to_owned(), None))
     }
 
     _ => {
@@ -298,7 +305,7 @@ fn process_list_call(
 ) -> Result<(Calcit, Option<Calcit>), CalcitErr> {
   let head = &xs[0];
   let (head_form, head_evaled) = preprocess_expr(head, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?;
-  let args = xs.skip(1)?;
+  let args = xs.drop_left();
   let def_name = grab_def_name(head);
 
   // println!(
@@ -312,7 +319,13 @@ fn process_list_call(
   //   }
   // );
 
-  match (head_form.to_owned(), head_evaled) {
+  // == Tips ==
+  // Macro from value: will be called during processing
+  // Func from value: for checking arity
+  // Keyword: transforming into keyword expression
+  // Syntax: handled directly during preprocessing
+  // Thunk: invalid here
+  match (&head_form, &head_evaled) {
     (Calcit::Keyword(..), _) => {
       if args.len() == 1 {
         let code = Calcit::List(TernaryTreeList::from(&[
@@ -335,17 +348,7 @@ fn process_list_call(
       }
     }
     (
-      Calcit::Macro {
-        name,
-        def_ns,
-        args: def_args,
-        body,
-        ..
-      },
       _,
-    )
-    | (
-      Calcit::Symbol { .. },
       Some(Calcit::Macro {
         name,
         def_ns,
@@ -354,42 +357,41 @@ fn process_list_call(
         ..
       }),
     ) => {
-      let mut current_values = args.to_owned();
+      let mut current_values = Box::new(args.to_owned());
 
       // println!("eval macro: {}", primes::CrListWrap(xs.to_owned()));
       // println!("macro... {} {}", x, CrListWrap(current_values.to_owned()));
 
       let code = Calcit::List(xs.to_owned());
-      let next_stack = extend_call_stack(call_stack, def_ns.to_owned(), name, StackKind::Macro, code, &args);
+      let next_stack = extend_call_stack(call_stack, def_ns.to_owned(), name.to_owned(), StackKind::Macro, code, &args);
 
       loop {
         // need to handle recursion
         // println!("evaling line: {:?}", body);
-        let body_scope = runner::bind_args(def_args.to_owned(), &current_values, &rpds::HashTrieMap::new_sync(), &next_stack)?;
-        let code = runner::evaluate_lines(&body, &body_scope, def_ns.to_owned(), &next_stack)?;
+        let body_scope = runner::bind_args(def_args, &current_values, &rpds::HashTrieMap::new_sync(), &next_stack)?;
+        let code = runner::evaluate_lines(body, &body_scope, def_ns.to_owned(), &next_stack)?;
         match code {
           Calcit::Recur(ys) => {
-            current_values = ys;
+            current_values = Box::new(ys.to_owned());
           }
           _ => {
             // println!("gen code: {} {}", code, &code.lisp_str());
-            let (final_code, v) = preprocess_expr(&code, scope_defs, file_ns, check_warnings, &next_stack)?;
-            return Ok((final_code, v));
+            return preprocess_expr(&code, scope_defs, file_ns, check_warnings, &next_stack);
           }
         }
       }
     }
     (Calcit::Syntax(name, name_ns), _) => match name {
       CalcitSyntax::Quasiquote => Ok((
-        preprocess_quasiquote(&name, name_ns, &args, scope_defs, file_ns, check_warnings, call_stack)?,
+        preprocess_quasiquote(name, name_ns.to_owned(), &args, scope_defs, file_ns, check_warnings, call_stack)?,
         None,
       )),
       CalcitSyntax::Defn | CalcitSyntax::Defmacro => Ok((
-        preprocess_defn(&name, name_ns, &args, scope_defs, file_ns, check_warnings, call_stack)?,
+        preprocess_defn(name, name_ns.to_owned(), &args, scope_defs, file_ns, check_warnings, call_stack)?,
         None,
       )),
       CalcitSyntax::CoreLet => Ok((
-        preprocess_call_let(&name, name_ns, &args, scope_defs, file_ns, check_warnings, call_stack)?,
+        preprocess_call_let(name, name_ns.to_owned(), &args, scope_defs, file_ns, check_warnings, call_stack)?,
         None,
       )),
       CalcitSyntax::If
@@ -398,14 +400,31 @@ fn process_list_call(
       | CalcitSyntax::MacroexpandAll
       | CalcitSyntax::Macroexpand1
       | CalcitSyntax::Reset => Ok((
-        preprocess_each_items(&name, name_ns, &args, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?,
+        preprocess_each_items(
+          name,
+          name_ns.to_owned(),
+          &args,
+          scope_defs,
+          file_ns.to_owned(),
+          check_warnings,
+          call_stack,
+        )?,
+        None,
+      )),
+      CalcitSyntax::Quote | CalcitSyntax::Eval | CalcitSyntax::HintFn => Ok((
+        preprocess_quote(name, name_ns.to_owned(), &args, scope_defs, file_ns.to_owned())?,
         None,
       )),
-      CalcitSyntax::Quote | CalcitSyntax::Eval | CalcitSyntax::HintFn => {
-        Ok((preprocess_quote(&name, name_ns, &args, scope_defs, file_ns.to_owned())?, None))
-      }
       CalcitSyntax::Defatom => Ok((
-        preprocess_defatom(&name, name_ns, &args, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?,
+        preprocess_defatom(
+          name,
+          name_ns.to_owned(),
+          &args,
+          scope_defs,
+          file_ns.to_owned(),
+          check_warnings,
+          call_stack,
+        )?,
         None,
       )),
     },
@@ -422,14 +441,14 @@ fn process_list_call(
         ..
       }),
     ) => {
-      check_fn_args(&f_args, &args, file_ns.to_owned(), f_name, def_name, check_warnings);
+      check_fn_args(f_args, &args, file_ns.to_owned(), f_name.to_owned(), def_name, check_warnings);
       let mut ys = Vec::with_capacity(args.len() + 1);
       ys.push(head_form);
       for a in &args {
         let (form, _v) = preprocess_expr(a, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?;
         ys.push(form);
       }
-      Ok((Calcit::List(TernaryTreeList::from(ys)), None))
+      Ok((Calcit::List(TernaryTreeList::from(&ys)), None))
     }
     (_, _) => {
       let mut ys = Vec::with_capacity(args.len() + 1);
@@ -438,14 +457,14 @@ fn process_list_call(
         let (form, _v) = preprocess_expr(a, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?;
         ys.push(form);
       }
-      Ok((Calcit::List(TernaryTreeList::from(ys)), None))
+      Ok((Calcit::List(TernaryTreeList::from(&ys)), None))
     }
   }
 }
 
 // detects arguments of top-level functions when possible
 fn check_fn_args(
-  defined_args: &Arc<Vec<Arc<str>>>,
+  defined_args: &[Arc<str>],
   params: &CalcitItems,
   file_ns: Arc<str>,
   f_name: Arc<str>,
@@ -536,7 +555,7 @@ pub fn preprocess_each_items(
   let mut xs: CalcitItems = TernaryTreeList::from(&[Calcit::Syntax(head.to_owned(), head_ns.to_owned())]);
   for a in args {
     let (form, _v) = preprocess_expr(a, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?;
-    xs = xs.push(form);
+    xs = xs.push_right(form);
   }
   Ok(Calcit::List(xs))
 }
@@ -564,7 +583,7 @@ pub fn preprocess_defn(
     ) => {
       let mut body_defs: HashSet<Arc<str>> = scope_defs.to_owned();
 
-      xs = xs.push(Calcit::Symbol {
+      xs = xs.push_right(Calcit::Symbol {
         sym: def_name.to_owned(),
         ns: def_name_ns.to_owned(),
         at_def: at_def.to_owned(),
@@ -577,7 +596,7 @@ pub fn preprocess_defn(
             sym, ns: def_ns, at_def, ..
           } => {
             check_symbol(sym, args, check_warnings);
-            zs = zs.push(Calcit::Symbol {
+            zs = zs.push_right(Calcit::Symbol {
               sym: sym.to_owned(),
               ns: def_ns.to_owned(),
               at_def: at_def.to_owned(),
@@ -596,12 +615,12 @@ pub fn preprocess_defn(
           }
         }
       }
-      xs = xs.push(Calcit::List(zs));
+      xs = xs.push_right(Calcit::List(zs));
 
       for (idx, a) in args.into_iter().enumerate() {
         if idx >= 2 {
           let (form, _v) = preprocess_expr(a, &body_defs, file_ns.to_owned(), check_warnings, call_stack)?;
-          xs = xs.push(form);
+          xs = xs.push_right(form);
         }
       }
       Ok(Calcit::List(xs))
@@ -676,11 +695,11 @@ pub fn preprocess_call_let(
       ))
     }
   };
-  xs = xs.push(binding);
+  xs = xs.push_right(binding);
   for (idx, a) in args.into_iter().enumerate() {
     if idx > 0 {
       let (form, _v) = preprocess_expr(a, &body_defs, file_ns.to_owned(), check_warnings, call_stack)?;
-      xs = xs.push(form);
+      xs = xs.push_right(form);
     }
   }
   Ok(Calcit::List(xs))
@@ -695,7 +714,7 @@ pub fn preprocess_quote(
 ) -> Result<Calcit, CalcitErr> {
   let mut xs: CalcitItems = TernaryTreeList::from(&[Calcit::Syntax(head.to_owned(), head_ns.to_owned())]);
   for a in args {
-    xs = xs.push(a.to_owned());
+    xs = xs.push_right(a.to_owned());
   }
   Ok(Calcit::List(xs))
 }
@@ -713,7 +732,7 @@ pub fn preprocess_defatom(
   for a in args {
     // TODO
     let (form, _v) = preprocess_expr(a, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?;
-    xs = xs.push(form.to_owned());
+    xs = xs.push_right(form.to_owned());
   }
   Ok(Calcit::List(xs))
 }
@@ -730,7 +749,7 @@ pub fn preprocess_quasiquote(
 ) -> Result<Calcit, CalcitErr> {
   let mut xs: CalcitItems = TernaryTreeList::from(&[Calcit::Syntax(head.to_owned(), head_ns.to_owned())]);
   for a in args {
-    xs = xs.push(preprocess_quasiquote_internal(
+    xs = xs.push_right(preprocess_quasiquote_internal(
       a,
       scope_defs,
       file_ns.to_owned(),
@@ -755,14 +774,14 @@ pub fn preprocess_quasiquote_internal(
         let mut xs: CalcitItems = TernaryTreeList::Empty;
         for y in ys {
           let (form, _) = preprocess_expr(y, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?;
-          xs = xs.push(form.to_owned());
+          xs = xs.push_right(form.to_owned());
         }
         Ok(Calcit::List(xs))
       }
       _ => {
         let mut xs: CalcitItems = TernaryTreeList::Empty;
         for y in ys {
-          xs = xs.push(preprocess_quasiquote_internal(y, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?.to_owned());
+          xs = xs.push_right(preprocess_quasiquote_internal(y, scope_defs, file_ns.to_owned(), check_warnings, call_stack)?.to_owned());
         }
         Ok(Calcit::List(xs))
       }
