@@ -3,26 +3,41 @@ import { overwriteComparator, initTernaryTreeMap } from "@calcit/ternary-tree";
 import { overwriteMapComparator } from "./js-map";
 
 import { CalcitRecord, fieldsEqual } from "./js-record";
-import { CalcitMap } from "./js-map";
+import { CalcitMap, CalcitSliceMap } from "./js-map";
 
 import { CalcitValue } from "./js-primes";
-import { CalcitList } from "./js-list";
+import { CalcitList, CalcitSliceList } from "./js-list";
 import { CalcitSet, overwriteSetComparator } from "./js-set";
 import { CalcitTuple } from "./js-tuple";
 
 // we have to inject cache in a dirty way in some cases
 const calcit_dirty_hash_key = "_calcit_cached_hash";
 
+let keywordIdx = 0;
+
 export class CalcitKeyword {
   value: string;
   cachedHash: Hash;
+  // use keyword for fast comparing
+  idx: number;
   constructor(x: string) {
     this.value = x;
+    this.idx = keywordIdx;
+    keywordIdx++;
     this.cachedHash = null;
   }
   toString() {
     return `:${this.value}`;
   }
+  cmp(other: CalcitKeyword): number {
+    if (this.idx < other.idx) {
+      return -1;
+    } else if (this.idx > other.idx) {
+      return 1;
+    } else {
+      return 0;
+    }
+  }
 }
 
 export class CalcitSymbol {
@@ -51,10 +66,10 @@ export class CalcitRecur {
 }
 
 export let isNestedCalcitData = (x: CalcitValue): boolean => {
-  if (x instanceof CalcitList) {
+  if (x instanceof CalcitList || x instanceof CalcitSliceList) {
     return x.len() > 0;
   }
-  if (x instanceof CalcitMap) {
+  if (x instanceof CalcitMap || x instanceof CalcitSliceMap) {
     return x.len() > 0;
   }
   if (x instanceof CalcitRecord) {
@@ -67,10 +82,10 @@ export let isNestedCalcitData = (x: CalcitValue): boolean => {
 };
 
 export let tipNestedCalcitData = (x: CalcitValue): string => {
-  if (x instanceof CalcitList) {
+  if (x instanceof CalcitList || x instanceof CalcitSliceList) {
     return "'[]...";
   }
-  if (x instanceof CalcitMap) {
+  if (x instanceof CalcitMap || x instanceof CalcitSliceMap) {
     return "'{}...";
   }
   if (x instanceof CalcitRecord) {
@@ -114,24 +129,24 @@ export let getStringName = (x: CalcitValue): string => {
 };
 
 /** returns -1 when not found */
-export function findInFields(xs: Array<string>, y: string): number {
+export function findInFields(xs: Array<CalcitKeyword>, y: CalcitKeyword): number {
   let lower = 0;
   let upper = xs.length - 1;
 
   while (upper - lower > 1) {
     let pos = (lower + upper) >> 1;
     let v = xs[pos];
-    if (y < v) {
+    if (y.idx < v.idx) {
       upper = pos - 1;
-    } else if (y > v) {
+    } else if (y.idx > v.idx) {
       lower = pos + 1;
     } else {
       return pos;
     }
   }
 
-  if (y == xs[lower]) return lower;
-  if (y == xs[upper]) return upper;
+  if (y === xs[lower]) return lower;
+  if (y === xs[upper]) return upper;
   return -1;
 }
 
@@ -148,6 +163,19 @@ export let kwd = (content: string) => {
   }
 };
 
+export let castKwd = (x: CalcitValue): CalcitKeyword => {
+  if (x instanceof CalcitKeyword) {
+    return x;
+  }
+  if (typeof x === "string") {
+    return kwd(x);
+  }
+  if (x instanceof CalcitSymbol) {
+    return kwd(x.value);
+  }
+  throw new Error(`Cannot cast this to keyword: ${x}`);
+};
+
 export var refsRegistry = new Map<string, CalcitRef>();
 
 let defaultHash_nil = valueHash("nil:");
@@ -189,7 +217,7 @@ let hashFunction = (x: CalcitValue): Hash => {
   }
 
   if (x instanceof CalcitKeyword) {
-    let h = mergeValueHash(defaultHash_keyword, x.value);
+    let h = mergeValueHash(defaultHash_keyword, x.idx);
     x.cachedHash = h;
     return h;
   }
@@ -232,6 +260,16 @@ let hashFunction = (x: CalcitValue): Hash => {
     }
     return base;
   }
+  if (x instanceof CalcitSliceList) {
+    let base = defaultHash_list;
+    // low-level code for perf
+    for (let idx = x.start; idx < x.end; idx++) {
+      let item = x.value[idx];
+      base = mergeValueHash(base, hashFunction(item));
+    }
+    x.cachedHash = base;
+    return base;
+  }
   if (x instanceof CalcitList) {
     let base = defaultHash_list;
     for (let item of x.items()) {
@@ -240,6 +278,15 @@ let hashFunction = (x: CalcitValue): Hash => {
     x.cachedHash = base;
     return base;
   }
+  if (x instanceof CalcitSliceMap) {
+    let base = defaultHash_map;
+    for (let [k, v] of x.pairs()) {
+      base = mergeValueHash(base, hashFunction(k));
+      base = mergeValueHash(base, hashFunction(v));
+    }
+    x.cachedHash = base;
+    return base;
+  }
   if (x instanceof CalcitMap) {
     let base = defaultHash_map;
     for (let [k, v] of x.pairs()) {
@@ -303,10 +350,10 @@ export let toString = (x: CalcitValue, escaped: boolean): string => {
   if (x instanceof CalcitKeyword) {
     return x.toString();
   }
-  if (x instanceof CalcitList) {
+  if (x instanceof CalcitList || x instanceof CalcitSliceList) {
     return x.toString();
   }
-  if (x instanceof CalcitMap) {
+  if (x instanceof CalcitMap || x instanceof CalcitSliceMap) {
     return x.toString();
   }
   if (x instanceof CalcitSet) {
@@ -354,14 +401,14 @@ export let to_js_data = (x: CalcitValue, addColon: boolean = false): any => {
     }
     return Symbol(x.value);
   }
-  if (x instanceof CalcitList) {
+  if (x instanceof CalcitList || x instanceof CalcitSliceList) {
     var result: any[] = [];
     for (let item of x.items()) {
       result.push(to_js_data(item, addColon));
     }
     return result;
   }
-  if (x instanceof CalcitMap) {
+  if (x instanceof CalcitMap || x instanceof CalcitSliceMap) {
     let result: Record<string, CalcitValue> = {};
     for (let [k, v] of x.pairs()) {
       var key = to_js_data(k, addColon);
@@ -378,8 +425,8 @@ export let to_js_data = (x: CalcitValue, addColon: boolean = false): any => {
   }
   if (x instanceof CalcitRecord) {
     let result: Record<string, CalcitValue> = {};
-    for (let idx in x.fields) {
-      result[x.fields[idx]] = to_js_data(x.values[idx]);
+    for (let idx = 0; idx < x.fields.length; idx++) {
+      result[x.fields[idx].value] = to_js_data(x.values[idx]);
     }
     return result;
   }
@@ -398,7 +445,7 @@ export let _$n_map_$o_get = function (xs: CalcitValue, k: CalcitValue) {
     throw new Error("map &get takes 2 arguments");
   }
 
-  if (xs instanceof CalcitMap) return xs.get(k);
+  if (xs instanceof CalcitMap || xs instanceof CalcitSliceMap) return xs.get(k);
 
   throw new Error("Does not support `&get` on this type");
 };
@@ -446,8 +493,8 @@ export let _$n__$e_ = (x: CalcitValue, y: CalcitValue): boolean => {
     }
     return false;
   }
-  if (x instanceof CalcitList) {
-    if (y instanceof CalcitList) {
+  if (x instanceof CalcitList || x instanceof CalcitSliceList) {
+    if (y instanceof CalcitList || y instanceof CalcitSliceList) {
       if (x.len() !== y.len()) {
         return false;
       }
@@ -463,8 +510,8 @@ export let _$n__$e_ = (x: CalcitValue, y: CalcitValue): boolean => {
     }
     return false;
   }
-  if (x instanceof CalcitMap) {
-    if (y instanceof CalcitMap) {
+  if (x instanceof CalcitMap || x instanceof CalcitSliceMap) {
+    if (y instanceof CalcitMap || y instanceof CalcitSliceMap) {
       if (x.len() !== y.len()) {
         return false;
       }
@@ -539,7 +586,7 @@ export let _$n__$e_ = (x: CalcitValue, y: CalcitValue): boolean => {
       if (x.values.length !== y.values.length) {
         return false;
       }
-      for (let idx in x.fields) {
+      for (let idx = 0; idx < x.fields.length; idx++) {
         if (!_$n__$e_(x.values[idx], y.values[idx])) {
           return false;
         }
